@@ -18,6 +18,7 @@
 from typing import Union, Sized, List, Optional, Tuple
 
 import numpy as np
+import pandas as pd
 from sklearn.utils.multiclass import unique_labels
 from sklearn.utils.validation import check_X_y
 
@@ -62,7 +63,11 @@ class LinearLMEProblem(LMEProblem):
                  order_of_objects: np.ndarray,
                  categorical_features: List[np.ndarray] = None,
                  answers=None,
-                 categorical_features_bootstrap_idx=None):
+                 categorical_features_bootstrap_idx=None,
+                 fe_columns=None,
+                 re_columns=None,
+                 fe_regularization_weights=None,
+                 re_regularization_weights=None):
 
         super(LinearLMEProblem, self).__init__()
 
@@ -85,6 +90,11 @@ class LinearLMEProblem(LMEProblem):
 
         self.categorical_features_bootstrap_idx = categorical_features_bootstrap_idx
 
+        self.fe_columns = fe_columns
+        self.re_columns = re_columns
+        self.fe_regularization_weights = fe_regularization_weights
+        self.re_regularization_weights = re_regularization_weights
+
     def __iter__(self):
         self.__iteration_pos = 0
         return self
@@ -114,6 +124,10 @@ class LinearLMEProblem(LMEProblem):
                  return_true_model_coefficients: bool = True,
                  seed: int = None,
                  generator_params: dict = None,
+                 chance_missing: float = 0.0,
+                 chance_outlier: float = 0.0,
+                 outlier_multiplier: float = 5.0,
+                 distribution="normal",
                  ):
         """
 
@@ -303,9 +317,14 @@ class LinearLMEProblem(LMEProblem):
             # generate ALL the data iteratively and
             # try to find features which don't give too much granular subdivision
             if num_features_to_generate > 0:
-                all_data_in_one_matrix = np.random.multivariate_normal(np.zeros(num_features_to_generate),
-                                                                       features_covariance_matrix,
-                                                                       num_objects)
+                if distribution == "normal":
+                    all_data_in_one_matrix = np.random.multivariate_normal(np.zeros(num_features_to_generate),
+                                                                           features_covariance_matrix,
+                                                                           num_objects)
+                elif distribution == "uniform":
+                    all_data_in_one_matrix = np.random.uniform(-2, 2, (num_objects, num_features_to_generate))
+                else:
+                    raise ValueError("Unknown distribution")
                 # add intercept to the left
                 all_data_in_one_matrix = np.hstack([np.ones(num_objects).reshape((-1, 1)), all_data_in_one_matrix])
             else:
@@ -354,6 +373,7 @@ class LinearLMEProblem(LMEProblem):
 
             group_continuous_features = all_data_in_one_matrix[np.ix_(group_idxs, continuous_features_idx)]
             group_fixed_features = group_continuous_features[:, fixed_features_idx]
+            # TODO: this one fails if there are no random effects. Fix it.
             group_random_features = group_continuous_features[:, random_features_idx]
 
             random_effects = None
@@ -379,6 +399,19 @@ class LinearLMEProblem(LMEProblem):
             group_errors = np.random.randn(group_size) * std
             group_answers = group_fixed_features.dot(beta) + group_random_features.dot(random_effects) + group_errors
 
+            # generate outliers and missing values after the answers are generated
+            missing_and_outliers_mask = np.random.choice([0, 1, outlier_multiplier],
+                                                         size=group_continuous_features.shape,
+                                                         p=[
+                                                             chance_missing,
+                                                             1 - chance_missing - chance_outlier,
+                                                             chance_outlier
+                                                         ])
+            group_continuous_features *= missing_and_outliers_mask
+            group_fixed_features = group_continuous_features[:, fixed_features_idx]
+            # TODO: this one fails if there are no random effects. Fix it.
+            group_random_features = group_continuous_features[:, random_features_idx]
+
             true_group_labels[start: start + group_size] = i
 
             random_effects_list.append((group_label, random_effects))
@@ -412,7 +445,7 @@ class LinearLMEProblem(LMEProblem):
         data['column_labels'] = all_columns_labels
 
         # We pivot the groups back to the original group division
-        generated_problem = LinearLMEProblem(**data).pivot(categorical_features_set=(0, ))
+        generated_problem = LinearLMEProblem(**data).pivot(categorical_features_set=(0,))
         generated_problem = generated_problem.to_x_y() if as_x_y else generated_problem
 
         if return_true_model_coefficients:
@@ -438,9 +471,13 @@ class LinearLMEProblem(LMEProblem):
     @staticmethod
     def from_x_y(x: np.ndarray,
                  y: Optional[np.ndarray] = None,
+                 columns: List[str] = None,
                  columns_labels: List[int] = None,
                  random_intercept: bool = True,
+                 fixed_intercept: bool = True,
                  add_group_as_categorical_feature=False,
+                 must_include_fe: List[bool] = None,
+                 must_include_re: List[bool] = None,
                  **kwargs):
         """
         Transforms matrices x (data) and y(answers) into an instance of LinearLMEProblem
@@ -459,6 +496,9 @@ class LinearLMEProblem(LMEProblem):
             There should be only one 0 in the list. If columns_labels is None then it's assumed that
             it is the first row of x.
 
+        fixed_intercept: bool, default = True
+            Whether to treat the intercept as a fixed featue
+
         random_intercept: bool, default = True
             Whether to treat the intercept as a random feature.
 
@@ -488,7 +528,7 @@ class LinearLMEProblem(LMEProblem):
         obs_std_idx = [i for i, label in enumerate(columns_labels) if label == 4]
         assert len(obs_std_idx) == 1, "There should be only one 4 in columns_labels"
         obs_std_idx = obs_std_idx[0]
-        assert all(x[:, obs_std_idx] != 0), "Errors' STDs can't be zero. Check for zeros in the respective column."
+        assert all(x[:, obs_std_idx] != 0), "Errors' STD can't be zero. Check for zeros in the respective column."
         fixed_features_idx = [i for i, t in enumerate(columns_labels) if t == 1 or t == 3]
         random_features_idx = [i for i, t in enumerate(columns_labels) if t == 2 or t == 3]
         # We include the groups column to the list of categorical features
@@ -497,6 +537,25 @@ class LinearLMEProblem(LMEProblem):
         num_categorical_features = len(categorical_features_idx)
         groups_labels = unique_labels(x[:, group_labels_idx])
 
+        fe_regularization_weights = [1] * (len(fixed_features_idx) + 1) if not must_include_fe \
+            else [int(not must_include_fe[-1])] + [int(not must_include_fe[i]) for i in fixed_features_idx]
+        re_regularization_weights = [1] * (len(random_features_idx) + 1) if not must_include_re \
+            else [int(not must_include_re[-1])] + [int(not must_include_re[i]) for i in random_features_idx]
+
+        if fixed_intercept & random_intercept:
+            intercept_label = [3]
+        elif random_intercept and not fixed_intercept:
+            intercept_label = [2]
+        elif fixed_intercept and not random_intercept:
+            intercept_label = [1]
+        else:
+            intercept_label = []
+
+        fe_columns = ((["intercept"] if fixed_intercept else []) + [columns[i] for i in
+                                                                    fixed_features_idx]) if columns else None
+        re_columns = ((["intercept"] if random_intercept else []) + [columns[i] for i in
+                                                                     random_features_idx]) if columns else None
+
         data = {
             'fixed_features': [],
             'random_features': [],
@@ -504,7 +563,11 @@ class LinearLMEProblem(LMEProblem):
             'answers': None if y is None else [],
             'obs_stds': [],
             'group_labels': groups_labels,
-            'column_labels': np.array([3 if random_intercept else 1] + columns_labels),
+            'column_labels': np.array(intercept_label + columns_labels),
+            'fe_columns': fe_columns,
+            're_columns': re_columns,
+            'fe_regularization_weights': fe_regularization_weights,
+            're_regularization_weights': re_regularization_weights,
         }
 
         order_of_objects = []
@@ -514,8 +577,11 @@ class LinearLMEProblem(LMEProblem):
             fixed_features = x[np.ix_(objects_idx, fixed_features_idx)]
             # add an intercept column plus real features
             # noinspection PyTypeChecker
-            data['fixed_features'].append(
-                np.concatenate((np.ones((fixed_features.shape[0], 1)), fixed_features), axis=1))
+            if fixed_intercept:
+                data['fixed_features'].append(
+                    np.concatenate((np.ones((fixed_features.shape[0], 1)), fixed_features), axis=1))
+            else:
+                data['fixed_features'].append(fixed_features)
             # same for random effects
             random_features = x[np.ix_(objects_idx, random_features_idx)]
             if random_intercept:
@@ -557,12 +623,18 @@ class LinearLMEProblem(LMEProblem):
         all_stds = np.concatenate(self.obs_stds, axis=0)
         untitled_data = np.zeros((all_fixed_features.shape[0], len(self.column_labels) - 1))
 
-        fixed_effects_counter = 1
-        random_intercept = self.column_labels[0] == 3
+        fixed_intercept = (self.column_labels[0] == 1) or (self.column_labels[0] == 3)
+        if fixed_intercept:
+            fixed_effects_counter = 1
+        else:
+            fixed_effects_counter = 0
+
+        random_intercept = (self.column_labels[0] == 2) or (self.column_labels[0] == 3)
         if random_intercept:
             random_effects_counter = 1
         else:
             random_effects_counter = 0
+
         categorical_features_counter = 0
 
         for i, label in enumerate(self.column_labels[1:]):
@@ -594,6 +666,60 @@ class LinearLMEProblem(LMEProblem):
         all_answers = all_answers[np.array(self.order_of_objects).argsort()]
         return data_with_column_labels, all_answers
 
+    @staticmethod
+    def from_dataframe(data: pd.DataFrame,
+                       fixed_effects: List[str],
+                       random_effects: List[str],
+                       groups: str,
+                       obs_std: str,
+                       target: str,
+                       must_include_fe: List[str],
+                       must_include_re: List[str]
+                       ):
+        if "intercept" in data.columns:
+            raise ValueError("intercept should not be in the data-frame")
+
+        for effect in fixed_effects + random_effects + [groups, obs_std, target] + must_include_fe + must_include_re:
+            if (effect != "intercept") and (effect not in data.columns):
+                raise ValueError(f"{effect} is not a column of the data-frame")
+
+        if not all(effect in fixed_effects for effect in must_include_fe):
+            raise ValueError("All elements from must_include_fe should also be in fixed_effects")
+        if not all(effect in random_effects for effect in must_include_re):
+            raise ValueError("All elements from must_include_re should also be in random_effects")
+
+        columns = []
+        column_labels = []
+
+        for effect in data.columns:
+            if (effect not in fixed_effects) and (effect not in random_effects):
+                continue
+            if (effect in fixed_effects) and (effect in random_effects):
+                column_labels.append(3)
+            elif effect in random_effects:
+                column_labels.append(2)
+            else:
+                column_labels.append(1)
+            columns.append(effect)
+
+        columns = [groups, obs_std] + columns
+        column_labels = [0, 4] + column_labels
+        must_include_fe_flags = [effect in must_include_fe for effect in columns] + \
+                                ["intercept" in must_include_fe]
+        must_include_re_flags = [effect in must_include_re for effect in columns] + \
+                                ["intercept" in must_include_re]
+
+        x = data[columns].to_numpy()
+        y = data[target].to_numpy()
+        return LinearLMEProblem.from_x_y(x=x,
+                                         y=y,
+                                         columns_labels=column_labels,
+                                         columns=columns,
+                                         fixed_intercept="intercept" in fixed_effects,
+                                         random_intercept="intercept" in random_effects,
+                                         must_include_fe=must_include_fe_flags,
+                                         must_include_re=must_include_re_flags)
+
     def pivot(self, categorical_features_set):
         """
         Get a transformed problem such that its groups labels are formed by all different combinations of
